/******************************************************************************
* C Include: Plate                                                            
*                                                                             
* File:                                                                       
*    plate.h                                                                  
*                                                                             
* Description:                                                                
*    Include file for the Plate functions.                                    
*                                                                             
* History:                                                                    
*    Created    : 15-JUN-93 by J. Doggett,                                    
*       for the Association of Universities for Research in Astronomy, Inc.   
*          at the Space Telescope Science Institute,                          
*       to facilitate the accessing of data from the CD-ROM set of the        
*          Digitized Sky Survey.                                              
*    Delivered  : 01-MAR-94 by J. Doggett,                                    
*       GetImage Version 1.0.                                                 
*                                                                             
*    Updated    : 30-APR-95 by J. Doggett,                                    
*       Added volume label field to plate structure.                          
*       Added volume label to arguments comment for Create_Plate function.    
*    Redelivered: 18-AUG-95 by J. Doggett,                                    
*       GetImage Version 1.2.                                                 
*                                                                             
*    Updated    : 12-SEP-96 by J. Doggett,                                    
*       Removed prototype for the now obsolete function                       
*          Plates_Disc_Is_Available.                                          
*       Removed disc argument from prototype for Create_Plate function.       
*    Updated    : 21-NOV-96 by J. Doggett,
*       Added calibrationFile to Plate to support calibration extraction.
*       Added digitizedSurvey, astronomicalSurvey, bandpass for future 
*          support of multi-survey GetImage interface.
*       Added PLATE_RECORD_COMMENT_CHARACTER and PLATE_RECORD_EOL_CHAR macros.
*       Renamed Read_Plate_Info to Read_Version_1_Plate_Info.
*       Added prototype for new function Read_Version_2_Plate_Info.
*    Updated    : 22-NOV-96 by J. Doggett,
*       Renamed Create_Plate function to Create_Empty_Plate and modified
*          prototype corresponding to functions change.
*       Removed source from plate structure because info in other fields added.
*    Redelivered: xx-xxx-xx by J. Doggett,                                    
*       GetImage Version 2.0x.                                                
*                                                                             
* Copyright (c) 1993-1996, Association of Universities for Research in        
* Astronomy, Inc.  All Rights Reserved.                                       
* Produced under National Aeronautics and Space Administration grant          
* NAG W-2166.                                                                 
******************************************************************************/
#ifndef __PLATE__

#include <stdio.h>

#include "generic.h"

/******************
* Plate constants *
******************/
#define __PLATE__

#define PLATE_RECORD_COMMENT_CHARACTER '#'
#define PLATE_RECORD_EOL_CHAR          '$'

/***************************
* Plate derived data types *
***************************/
struct Plate_Info /* Information about a plate */

/* *** Plate Include File ***                                              2 */


{
   float  plateInfoVersion      ; /* Corresponds to place list file version */
   char   Id    [ 5 ]           ;
   char   Region[ 7 ]           ;
   double Center_RA             ; /* In radians */
   double Center_Dec            ; /* In radians */
   double Epoch                 ;
   char   volumeLabel[24]       ; /* Full volume label of disc where the plate
                                   * may be found.
                                   */
   char   digitizedSurvey[5]    ; /* Digitized survey with which plate is
                                   *  associated.
                                   */
   char   astronomicalSurvey[51]; /* Astronomical survey with which plate is
                                   *  associated.
                                   */
   char   bandpass[16]          ; /* Generic bandpass associated with plate.*/
   char   calibrationFile[81]   ; /* FITS file with calibration data. */
};
typedef struct Plate_Info * PLATE     ;

/******************************
* Plate functions' prototypes *
******************************/
int   Compare_Plates           (/* PLATE   Plate_1             ,
				   PLATE   Plate_2              */);

PLATE Create_Empty_Plate       (/*                              */);

void  Delete_Plate             (/* PLATE   Plate                */);

BOOL  Plate_Directory_Exists   (/* STRING  Plate_Name          ,
				   STRING  Image_Root_Directory */);

PLATE Read_Version_1_Plate_Info(/* FILE   *InputFile            */);

PLATE Read_Version_2_Plate_Info(/* FILE   *InputFile            */);

#endif
